import { createHash } from "crypto"
import { emptyDir, readJson, remove } from "fs-extra-p"
import isCi from "is-ci"
import { tmpdir } from "os"
import * as path from "path"
import { deleteOldElectronVersion, downloadAllRequiredElectronVersions } from "./downloadElectron"

const rootDir = path.join(__dirname, "../../..")

const util = require(`${rootDir}/packages/builder-util/out/util`)
const isEmptyOrSpaces = util.isEmptyOrSpaces

const baseDir = process.env.ELECTRON_BUILDER_TMP_DIR || (process.platform === "darwin" && !require("is-ci") ? "/tmp" : tmpdir())
const TEST_TMP_DIR = path.join(baseDir, `et-${createHash("md5").update(__dirname).digest("hex")}`)

runTests()
  .catch(error => {
    console.error(error.stack || error)
    process.exit(1)
  })

async function runTests() {
  process.env.BABEL_JEST_SKIP = "true"

  if (process.env.CIRCLECI) {
    await emptyDir(TEST_TMP_DIR)
  }
  else {
    await Promise.all([
      deleteOldElectronVersion(),
      downloadAllRequiredElectronVersions(),
      emptyDir(TEST_TMP_DIR),
    ])
  }

  const testFiles = process.env.TEST_FILES

  const testPatterns: Array<string> = []
  if (!isEmptyOrSpaces(testFiles)) {
    testPatterns.push(...testFiles!!.split(","))
  }
  else if (!isEmptyOrSpaces(process.env.CIRCLE_NODE_INDEX)) {
    const circleNodeIndex = parseInt(process.env.CIRCLE_NODE_INDEX!!, 10)
    if (circleNodeIndex === 0) {
      testPatterns.push("debTest")
      testPatterns.push("fpmTest")
      testPatterns.push("winPackagerTest")
      testPatterns.push("winCodeSignTest")
      testPatterns.push("squirrelWindowsTest")
      testPatterns.push("nsisUpdaterTest")
      testPatterns.push("macArchiveTest")
      testPatterns.push("macCodeSignTest")
      testPatterns.push("extraMetadataTest")
      testPatterns.push("HoistedNodeModuleTest")
      testPatterns.push("configurationValidationTest")
      testPatterns.push("webInstallerTest")
    }
    else if (circleNodeIndex === 1) {
      testPatterns.push("oneClickInstallerTest")
    }
    else if (circleNodeIndex === 2) {
      testPatterns.push("snapTest")
      testPatterns.push("macPackagerTest")
      testPatterns.push("linuxPackagerTest")
      testPatterns.push("msiTest")
      testPatterns.push("ignoreTest")
      testPatterns.push("mainEntryTest")
      testPatterns.push("ArtifactPublisherTest")
      testPatterns.push("RepoSlugTest")
      testPatterns.push("portableTest")
      testPatterns.push("globTest")
      testPatterns.push("BuildTest")
      testPatterns.push("linuxArchiveTest")
    }
    else {
      testPatterns.push("PublishManagerTest")
      testPatterns.push("assistedInstallerTest")
      testPatterns.push("filesTest")
    }
    console.log(`Test files for node ${circleNodeIndex}: ${testPatterns.join(", ")}`)
  }

  process.env.TEST_TMP_DIR = TEST_TMP_DIR

  const rootDir = path.join(__dirname, "..", "..", "..")

  const config = (await readJson(path.join(rootDir, "package.json"))).jest
  // use custom cache dir to avoid https://github.com/facebook/jest/issues/1903#issuecomment-261212137
  config.cacheDirectory = process.env.JEST_CACHE_DIR || "/tmp/jest-electron-builder-tests"
  // no need to transform — compiled before
  config.transform = {}
  config.transformIgnorePatterns = [".*"]
  config.bail = process.env.TEST_BAIL === "true"

  let runInBand = false
  const scriptArgs = process.argv.slice(2)

  const testPathIgnorePatterns = config.testPathIgnorePatterns
  if (scriptArgs.length > 0) {
    for (const scriptArg of scriptArgs) {
      console.log(`custom opt: ${scriptArg}`)
      if ("runInBand" === scriptArg) {
        runInBand = true
      }
      else if (scriptArg.includes("=")) {
        const equalIndex = scriptArg.indexOf("=")
        const envName = scriptArg.substring(0, equalIndex)
        let envValue = scriptArg.substring(equalIndex + 1)
        if (envValue === "isCi") {
          envValue = isCi ? "true" : "false"
        }

        process.env[envName] = envValue
        console.log(`Custom env ${envName}=${envValue}`)

        if (envName === "ALL_TESTS" && envValue === "false") {
          config.cacheDirectory += "-basic"
        }
      }
      else if (scriptArg.startsWith("skip")) {
        if (!isCi) {
          const suffix = scriptArg.substring("skip".length)
          switch (scriptArg) {
            case "skipArtifactPublisher": {
              testPathIgnorePatterns.push("[\\/]{1}ArtifactPublisherTest.js$")
              config.cacheDirectory += `-${suffix}`
            }
              // noinspection TsLint
              break

            default:
              throw new Error(`Unknown opt ${scriptArg}`)
          }
        }
      }
      else {
        config[scriptArg] = true
      }
    }
  }

  const jestOptions: any = {
    verbose: true,
    updateSnapshot: process.env.UPDATE_SNAPSHOT === "true",
    config,
    runInBand,
    projects: [rootDir],
  }

  if (testPatterns.length > 0) {
    jestOptions.testPathPattern = testPatterns
      .map(it => it.endsWith(".js") || it.endsWith("*") ? it : `${it}\\.js$`)
  }
  if (process.env.CIRCLECI != null || process.env.TEST_JUNIT_REPORT === "true") {
    jestOptions.reporters = ["default", "jest-junit"]
  }

  // console.log(JSON.stringify(jestOptions, null, 2))

  const testResult = await require("jest-cli").runCLI(jestOptions, jestOptions.projects)
  const exitCode = testResult.results == null || testResult.results.success ? 0 : testResult.globalConfig.testFailureExitCode
  if (isCi) {
    process.exit(exitCode)
  }

  await remove(TEST_TMP_DIR)
  process.exitCode = exitCode
  if (testResult.globalConfig.forceExit) {
    process.exit(exitCode)
  }
}